@@ -8,15 +8,17 @@ package App::MatrixTool;
 use strict;
 use warnings;
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 use Getopt::Long qw( GetOptionsFromArray );
 use Future;
+use List::Util 1.29 qw( pairmap );
 use MIME::Base64 qw( encode_base64 );
 use Module::Pluggable::Object;
 use Module::Runtime qw( use_package_optimistically );
 use Scalar::Util qw( blessed );
 use Socket qw( getnameinfo AF_INET AF_INET6 AF_UNSPEC NI_NUMERICHOST NI_NUMERICSERV );
+use Struct::Dumb qw( readonly_struct );
 
 my $opt_parser = Getopt::Long::Parser->new(
    config => [qw( require_order )],
@@ -40,6 +42,35 @@ C<App::MatrixTool::Command::> namespace, for details on specific commands.
 
 =cut
 
+readonly_struct ArgSpec => [qw( name print_name optional )];
+sub ARGSPECS
+{
+   map {
+      my $name = $_;
+      my $optional = $name =~ s/\?$//;
+
+      ( my $print_name = uc $name ) =~ s/_/-/g;
+
+      ArgSpec( $name, $print_name, $optional )
+   } shift->ARGUMENTS;
+}
+
+readonly_struct OptSpec => [qw( name print_name shortname getopt description )];
+sub OPTSPECS
+{
+   pairmap {
+      my $getopt = $a;
+
+      my $shortname;
+      $a =~ s/^(.)\|// and $shortname = $1;
+
+      my $printname = "$a";  # stringify to allocate a new SVtPV - RT95409
+      $a =~ s/-/_/g;
+
+      OptSpec( $a, $printname, $shortname, $getopt, $b )
+   } shift->OPTIONS;
+}
+
 sub new
 {
    my $class = shift;
@@ -54,6 +85,18 @@ sub sock_family
    return AF_UNSPEC;
 }
 
+sub _pkg_for_command
+{
+   my $self = shift;
+   my ( $cmd ) = @_;
+
+   # Allow hyphens in command names
+   $cmd =~ s/-/_/g;
+
+   my $pkg = "App::MatrixTool::Command::$cmd";
+   use_package_optimistically( $pkg );
+}
+
 sub run
 {
    my $self = shift;
@@ -68,27 +111,38 @@ sub run
    ) or return 1;
 
    my $cmd = @args ? shift @args : "help";
-   # Allow hyphens in command names
-   $cmd =~ s/-/_/g;
-
-   my $pkg = "App::MatrixTool::Command::$cmd";
-   use_package_optimistically( $pkg );
 
+   my $pkg = $self->_pkg_for_command( $cmd );
    $pkg->can( "new" ) or
       return $self->error( "No such command '$cmd'" );
 
+   my @argvalues;
+
    if( $pkg->can( "OPTIONS" ) ) {
-      my %optdefs = $pkg->OPTIONS;
       my %optvalues;
 
       GetOptionsFromArray( \@args,
-         map { $_ => \$optvalues{ $optdefs{$_} } } keys %optdefs,
+         map { $_->getopt => \$optvalues{ $_->name } } $pkg->OPTSPECS
       ) or exit 1;
 
-      unshift @args, \%optvalues;
+      push @argvalues, \%optvalues;
    }
 
-   my $ret = $pkg->new( %global_opts )->run( @args );
+   my @argspecs = $pkg->ARGSPECS;
+   while( @argspecs ) {
+      my $spec = shift @argspecs;
+
+      if( !@args ) {
+         last if $spec->optional;
+
+         return $self->error( "Required argument '${\ $spec->print_name }' missing" );
+      }
+
+      push @argvalues, shift @args;
+   }
+   @args and return $self->error( "Found extra arguments" );
+
+   my $ret = $pkg->new( %global_opts )->run( @argvalues );
    $ret = $ret->get if blessed $ret and $ret->isa( "Future" );
    $ret //= 0;
 
@@ -202,6 +256,10 @@ use base qw( App::MatrixTool );
 use List::Util qw( max );
 
 use constant DESCRIPTION => "Display help information about commands";
+use constant ARGUMENTS => ( "command?" );
+
+use Struct::Dumb qw( readonly_struct );
+readonly_struct CommandSpec => [qw( name description argspecs optspecs )];
 
 sub commands
 {
@@ -213,28 +271,38 @@ sub commands
    my @commands;
 
    foreach my $module ( sort $mp->plugins ) {
+      $module->can( "DESCRIPTION" ) or next;
+
       my $cmd = $module;
       $cmd =~ s/^App::MatrixTool::Command:://;
       $cmd =~ s/_/-/g;
 
-      push @commands, [ $cmd, $module->DESCRIPTION ];
+      push @commands, CommandSpec(
+         $cmd,
+         $module->DESCRIPTION,
+         [ $module->ARGSPECS ],
+         $module->can( "OPTIONS" ) ? [ $module->OPTSPECS ] : undef,
+      );
    }
 
    return @commands;
 }
 
-sub run
-{
-   my $self = shift;
-
-   $self->output( <<'EOF' );
-matrixtool [<global options...>] <command> [<command-options...>]
-
+my $GLOBAL_OPTS = <<'EOF';
 Global options:
    -4 --inet4             Use only IPv4
    -6 --inet6             Use only IPv6
       --print-request     Print sent HTTP requests in full
       --print-response    Print received HTTP responses in full
+EOF
+
+sub help_summary
+{
+   my $self = shift;
+
+   $self->output( <<'EOF' . $GLOBAL_OPTS );
+matrixtool [<global options...>] <command> [<command options...>]
+
 EOF
 
    my @commands = $self->commands;
@@ -242,10 +310,79 @@ EOF
    my $namelen = max map { length $_->[0] } @commands;
 
    $self->output( "Commands:\n" .
-      join "\n", map { sprintf "  %-*s    %s", $namelen, @$_ } @commands
+      join "\n", map { sprintf "  %-*s    %s", $namelen, $_->name, $_->description } @commands
    );
 }
 
+sub _argdesc
+{
+   shift;
+   my ( $argspec ) = @_;
+   my $name = $argspec->print_name;
+   return $argspec->optional ? "[$name]" : $name;
+}
+
+sub _optdesc
+{
+   my ( $optspec, $namelen ) = @_;
+
+   my $shortname = $optspec->shortname;
+
+   join "",
+      ( defined $shortname ? "-$shortname " : "   " ),
+      sprintf( "--%-*s", $namelen, $optspec->print_name ),
+      "    ",
+      $optspec->description,
+}
+
+sub help_detailed
+{
+   my $self = shift;
+   my ( $cmd ) = @_;
+
+   my $pkg = $self->_pkg_for_command( $cmd );
+   $pkg->can( "new" ) or
+      return $self->error( "No such command '$cmd'" );
+
+   my @argspecs = $pkg->ARGSPECS;
+
+   $self->output( join " ", "matrixtool",
+      "[<global options...>]",
+      $cmd,
+      ( map $self->_argdesc($_), @argspecs ),
+      "[<command options...>]"
+   );
+   $self->output( $pkg->DESCRIPTION );
+   $self->output();
+
+   $self->output( $GLOBAL_OPTS );
+
+   if( $pkg->can( "OPTIONS" ) ) {
+      my @optspecs = $pkg->OPTSPECS;
+
+      my $namelen = max map { length $_->name } @optspecs;
+
+      $self->output( "Options:" );
+      foreach my $optspec ( sort { $a->name cmp $b->name } @optspecs ) {
+         $self->output( "   " . _optdesc( $optspec, $namelen ) );
+      }
+      $self->output();
+   }
+}
+
+sub run
+{
+   my $self = shift;
+   my ( $cmd ) = @_;
+
+   if( defined $cmd ) {
+      $self->help_detailed( $cmd );
+   }
+   else {
+      $self->help_summary;
+   }
+}
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
